#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "readmatrix.h"

#define _REENTRANT
#include<pthread.h>


int numproc;


void *sub(void *);



typedef struct threaddata threaddata;

struct threaddata {
	float **arr;
	int start;
	int stop;
	int size;
	int *gotresult;
	pthread_cond_t *haveresult;
	int *readdata;
	pthread_cond_t *havereaddata;
	pthread_mutex_t *mutex;
	int number;
	threaddata *data;

	float maxchange;
};

int checkmaxchange(threaddata *data,float threshold,int numproc);


int main(int argc, char *argv[])
{
	char *filename;
	
	int size;
	float threshold;
	
	float **arr;

	int i,j;
	pthread_t *threads;
	threaddata *data;

	if (argc != 3) {
		printf("Usage: prog filename numproc\n");
		return 1;
	}
	filename = argv[1];
	numproc = atoi(argv[2]);
	readmatrix(filename, &arr, &size,&threshold);
	
	printf("\n%s %i\n",filename,numproc);
	
	if (numproc > size) {
		numproc = size;
	}

	threads = (pthread_t *)malloc(numproc*sizeof(pthread_t));
	data = (threaddata *)malloc(numproc*sizeof(threaddata));

	printf("This is main thread\n");

	
	for (i=0;i<numproc;i++) {
		data[i].arr = arr;
		data[i].start = size/numproc*i;
		data[i].stop = size/numproc*(i+1);
		data[i].size = size;
		data[i].number = i;
		data[i].data = data;
		data[i].maxchange = 0;
		if (i == numproc - 1 )
			data[i].stop += size % numproc;
		data[i].gotresult = (int *)malloc(size*sizeof(int));
		data[i].mutex = (pthread_mutex_t *)malloc(size*sizeof(pthread_mutex_t));
		data[i].haveresult = (pthread_cond_t *)malloc(size*sizeof(pthread_cond_t));
		for (j=0;j<size;j++) {
			pthread_mutex_init(&(data[i].mutex[j]), NULL);
			pthread_cond_init(&(data[i].haveresult[j]),NULL);
			data[i].gotresult[j] = 0;
		}

	}
	
	do {
		for (i=0;i<numproc;i++) {
			data[i].maxchange = 0;
			if (pthread_create(&threads[i], NULL, sub, &data[i] ))
			         fprintf(stderr, "Can't create thread\n"), exit(1);
			}
	
		printf("Main thread still\n");

		for (i=0;i<numproc;i++) {
			if (pthread_join(threads[i], NULL)!=0) {perror("Join error\n"); exit(-1);}
	        }
	} while (checkmaxchange(data,threshold,numproc) == 1);

	printf("Result is\n");
        for (i=0;i < size; i++) {
               for (j=0; j< size; j++)
                       printf("%3.2f ",arr[i][j]);
               printf("\n");
       }
			
}

int checkmaxchange(threaddata *data,float threshold,int numproc)
{
	int i;
	float max;

	max = 0;
	for (i=0;i<numproc;i++) {
		if (data[i].maxchange > max)
			max = data[i].maxchange;
	}
	if (max < threshold) {
		fprintf(stderr,"max change is %f, threshold is %f, finishing\n", max,threshold);
		return 0;
	} else {
		fprintf(stderr,"max change is %f, threshold is %f, continuing\n",max,threshold);
		return 1;
	}
}
			

void * sub(void *arg)
{
	threaddata *data = (threaddata *) arg;
	int me;
	int i,j;
	int start, stop;
	float old, change;
	pthread_mutex_t *mutex;
	pthread_cond_t *condresult;
	pthread_cond_t *condread;
	float up, down, left, right;

	me = pthread_self();

	printf("This is thread %i\n", me);

	
	for (i=1;i<(data->size) -1; i++) {
	       fprintf(stderr,"Thread %i locking own data...",data->number);
	       pthread_mutex_lock(&((data->mutex)[i]));
	       fprintf(stderr,"... success!\n");
	       
	       	start = data->start;
		stop = data->stop;
		if (data->start == 0)
			start++;
		if (data->stop == data->size)
			stop--;
		
                for (j=start;j<stop;j++) {
                       left = right = up = down = 0;
			
		       mutex = &((data->data)[(data->number)-1].mutex[i]);
		       condread =  &((data->data)[(data->number)-1].havereaddata[i]);
                       condresult = &((data->data)[(data->number)-1].haveresult[i]);
/*		       pthread_mutex_lock(mutex);
		       if ((data->data)[(data->number)-1].readdata == 0)
			        pthread_cond_wait(condread,mutex);
		       pthread_mutex_unlock(mutex);*/
		       
		       if (j == data->start) {
			       fprintf(stderr,"Thread %i locking data of previous thread...",data->number);
			       pthread_mutex_lock(mutex);
			       fprintf(stderr,"... success!\n");
			       if ((data->data)[(data->number)-1].gotresult[i] == 0) {
				       fprintf(stderr,"Data from previous thread not ready, thread %i waiting on cond var...",data->number);
				       pthread_cond_wait(condresult,mutex);
				       fprintf(stderr,"success!\n");
				}
			       left = data->arr[i][data->start-1];
			       fprintf(stderr,"Thread %i unlocking data of previous thread...\n",data->number);
			       pthread_mutex_unlock(mutex);
		       } else
			       left = data->arr[i][j-1];
			up = data->arr[i-1][j];
			down = data->arr[i+1][j];
			right = data->arr[i][j+1];
			old = data->arr[i][j];
		       data->arr[i][j] = (left + right + up + down) / 4;
		       change = old - data->arr[i][j];
		       if (change > data->maxchange)
		       		data->maxchange = change;
		} 
		(data->gotresult)[i] = 1;
		fprintf(stderr,"Thread %i calculated the row %i, broadcasts the condition and unlocks the data\n",data->number,i);
		pthread_cond_broadcast(&(data->haveresult[i]));
		pthread_mutex_unlock(&(data->mutex[i]));
		
	}
	printf("thread %i exiting\n",data->number);
	return NULL;
}
		       
				       
